#define GAMEX_MAIN

#include <windows.h>
#include <process.h>
#include <stdio.h>

#include "gamex.hpp"
#include "LogFile.hpp"
#include "Console.hpp"
#include "Graph.hpp"
#include "Button.hpp"
#include "SerialPort.hpp"
#include "Point.hpp"
#include "Elbow.hpp"
#include "Wrist.hpp"
#include "Body.hpp"

#define PI	3.14159f

int width = 512;
int	height = 768;
char buffer[128];

// Camera for 3D projections
CameraX cam;						
Vector3DF view;
Vector3DF axisView;
char txt_x[] = "x";
char txt_y[] = "y";
char txt_z[] = "z";

// GUI
HWND hWnd;
Console *console;
Graph *accGraph[3], *velGraph[3], *posGraph[3], *angGraph[3];
bool showAllGraphs = false, graphsDisabled = true;
Button *startLogButton, *saveLogButton, *runLogButton, *resetButton;

// Motion Capture
Wrist *wrist;
Elbow *elbow;
Point *shoulder;
Body *body;

// Serial Port Stuff
SerialPort serialPort;
BYTE ReadBuffer[255];
DWORD BytesRead;
HANDLE hDataMutex;
bool dataChanged = true, dataChanged_c = true;
Vector3DF accOne;
Vector3DF accTwo;
UINT b[9] = {0,0,0,0,0,0,0,0,0};
int b_ptr = 0;

// File Logging
LogFile logFile;
bool logSerialToFile = false, readFromLogFile = false;
bool doneReadingLog = false;
OPENFILENAME ofn;
char szFileName[128];
BYTE readBuf[128];

void GameInit (void)
{
	// ------------- GAMEX BEGIN ---------------
	GameX.Initialize ("Motion Capture", RUN_NOCONFIRMQUIT | VIDEO_16BIT
		| VIDEO_WINDOWED | RUN_BACKGROUND | RUN_AUTOSHOWINFO, 1280, 768);

	view.x = 0;	view.y = 0.5f; view.z = 250;
	axisView.x = 0; axisView.y = 0; axisView.z = 60;

	cam.SetWindow (0, 0, 512, 768);			// Initialize Camera
	cam.SetToPosition( Vector3DF(0,0,0) );	// Set camera to look at the origin
	cam.SetFromAngles( view );
	GameX.SetDrawCam(&cam);					// Tell GameX to draw using the camera
	// ------------- GAMEX END ---------------


	// ------------- GUI BEGIN ---------------
	startLogButton = new Button("Start Log", 10, 718, 100, 758);
	saveLogButton = new Button("Save Log", 110, 718, 200, 758);
	runLogButton = new Button("Run Log", 210, 718, 300, 758);
	resetButton = new Button("Reset", 310, 718, 400, 758);

	saveLogButton->Disable();
	
	hWnd = GameX.GetWindow();
	console = new Console("consoleLog.txt", 10,10,width-10,height-10);
	console->Printfn("Motion Capture Modelling System");
	console->Printfn("Debug Interface");

	accGraph[0] = new Graph("avg_acc1.x", -40, 40, width+0,   0, width+240, 190);
	accGraph[1] = new Graph("acc1.y", -40, 40, width+256, 0, width+496, 190);
	accGraph[2] = new Graph("acc1.z", -40, 40, width+512, 0, width+752, 190);
	velGraph[0] = new Graph("vel1.x", -4, 4, width+0,   192, width+240, 382);
	velGraph[1] = new Graph("vel1.y", -0.6, 0.6, width+256, 192, width+496, 382);
	velGraph[2] = new Graph("vel1.z", -0.6, 0.6, width+512, 192, width+752, 382);
	posGraph[0] = new Graph("pos1.x", -50, 50, width+0,   384, width+240, 574);
	posGraph[1] = new Graph("pos1.y", -50, 50, width+256, 384, width+496, 574);
	posGraph[2] = new Graph("pos1.z", -50, 50, width+512, 384, width+752, 574);
	angGraph[0] = new Graph("psi1",   -PI, PI,   width+0,   576, width+240, 766);
	angGraph[1] = new Graph("phi1",   -PI, PI, width+256, 576, width+496, 766);
	angGraph[2] = new Graph("theta1", -PI, PI, width+512, 576, width+752, 766);
	// ------------- GUI END ---------------


	// ------------- MOTION CAPTURE BEGIN ---------------
	shoulder = new Point( Vector3DF(0,0,0), 0, NULL, &cam);
	elbow = new Elbow( Vector3DF(0,0,-30), 30, shoulder, &cam);
	wrist = new Wrist( Vector3DF(0,-60,-30), 30, elbow, &cam);	
	body = new Body();
	// ------------- MOTION CAPTURE END ---------------


	// ------------- SERIAL PORT BEGIN ---------------
	if (serialPort.Init("COM1") == S_OK)
	{
		console->Printfn("COM1 Configured Correctly.");
	}
	else
	{
		// get error and display it
		char lpMsgBuf[100];
		serialPort.LastError(lpMsgBuf);
		console->Printfn("COM1 Configuration Error: %s", lpMsgBuf );
	}
	// ------------- SERIAL PORT END ---------------

	// ------------- LOG FILE START ---------------
	ofn.lStructSize = sizeof(OPENFILENAME); 
	ofn.hwndOwner = hWnd; 
	ofn.lpstrFilter = NULL; 
	ofn.lpstrFile= szFileName; 
	ofn.nMaxFile = sizeof(szFileName)/ sizeof(*szFileName);
	ofn.lpstrFileTitle = NULL; 
	ofn.nMaxFileTitle = 0; 
	ofn.lpstrInitialDir = (LPSTR)NULL; 
	ofn.Flags = OFN_SHOWHELP | OFN_OVERWRITEPROMPT | OFN_CREATEPROMPT; 
	ofn.lpstrTitle = NULL; 
	// ------------- LOG FILE END ---------------

	wrist->Reset();

}

float wristPsi = 0;
float elbowPhi = 0;

void GameRun (void)
{
	DWORD dwWaitResult; 

	if (readFromLogFile)
	{
		BytesRead = logFile.ReadLine(ReadBuffer);
		if (logFile.IsLogFinished())
		{
			logFile.Close();
			doneReadingLog = true;
			readFromLogFile = false;
			startLogButton->Enable();
			runLogButton->Enable();
		}
	}
	else if( serialPort.Read(ReadBuffer, &BytesRead) == E_FAIL )
	{
		// get error and display it
		char lpMsgBuf[100];
		serialPort.LastError(lpMsgBuf);

		console->Printfn("Serial Data Read Error: %s", lpMsgBuf );
	}

	for(int i=0; i<BytesRead; i++)
	{
		// if buffer is 0x01
		if( ReadBuffer[i] == 0x01 )
			b_ptr = 0;
		else
			b[b_ptr++] = ReadBuffer[i];

		// if you have a full data packet
		if( b_ptr == 8 )
		{
			b_ptr = 0;

			b[7] = ReadBuffer[i];

			// unpack the packet into values
			accOne.x = (float)b[0] + (float)( (b[6]&0xC0)<<2 );
			accOne.y = (float)b[1] + (float)( (b[6]&0x30)<<4 );
			accOne.z = (float)b[2] + (float)( (b[6]&0x0C)<<6 );

			accTwo.x = (float)b[3] + (float)( (b[6]&0x03)<<8 );
			accTwo.y = (float)b[4] + (float)( (b[7]&0xC0)<<2 );
			accTwo.z = (float)b[5] + (float)( (b[7]&0x30)<<4 );

			// display the data
			console->Printfn( "%f, %f, %f, %f, %f, %f", accOne.x, accOne.y, accOne.z, accTwo.x, accTwo.y, accTwo.z );

			elbow->Run(accTwo);
			wrist->Run(accOne);

			body->RotateElbow( elbow->phi, elbow->theta );
			body->RotateWrist( wrist->phi, wrist->psi);
			//body->RotateElbow( elbowPhi, 0 );
			//body->RotateWrist( wristPsi, PI/4);
			//elbowPhi += (PI/1920);
			//wristPsi += (PI/480);

			if (!graphsDisabled)
			{
				accGraph[0]->AddPoint(wrist->x_avg);//acc.x);
				accGraph[1]->AddPoint(wrist->acc.y);
				accGraph[2]->AddPoint(wrist->acc.z);

				velGraph[0]->AddPoint(wrist->vel.x);
				velGraph[1]->AddPoint(wrist->vel.y);
				velGraph[2]->AddPoint(wrist->vel.z);

				posGraph[0]->AddPoint(wrist->pos.x);
				posGraph[1]->AddPoint(wrist->pos.y);
				posGraph[2]->AddPoint(wrist->pos.z);

				angGraph[0]->AddPoint(wrist->psi);
				angGraph[1]->AddPoint(wrist->phi);
				angGraph[2]->AddPoint(wrist->theta);
			}
		}
	}

	if (logSerialToFile)
	{
		for (int i = 0; i < BytesRead; i++)
			logFile.Printf("%d ", ReadBuffer[i]);
		if (BytesRead > 0)
			logFile.Printf("\n");
	}

	if (startLogButton->Run())
	{
		logFile.OpenForWriting("accLogFile_do_not_touch.txt");
		logSerialToFile = true;
		doneReadingLog = false;
		saveLogButton->Enable();
	}
	if (saveLogButton->Run())
	{
		BOOL saveOK = GetSaveFileName(&ofn);
		if (saveOK)
		{
			logFile.Close();
			logSerialToFile = false;
			doneReadingLog = false;
			saveLogButton->Disable();
			CopyFile("accLogFile_do_not_touch.txt", ofn.lpstrFile, FALSE);
		}	
	}
	if (runLogButton->Run() && !logSerialToFile)
	{
		BOOL openOK = GetOpenFileName(&ofn);
		if (openOK)
		{
			logFile.OpenForReading(ofn.lpstrFile);
			readFromLogFile = true;
			doneReadingLog = false;
			startLogButton->Disable();
			saveLogButton->Disable();
			runLogButton->Disable();
		}	
	}
	if (resetButton->Run())
	{
		wrist->Reset();
		elbow->Reset();
		body->Reset();
	}

	// move camera view about the origin on right button down
	if( GameX.IsMouseDown(MOUSE_RIGHT) ){

		view.x += GameX.GetMouseDX();
		view.y += GameX.GetMouseDY();
		view.z += GameX.GetMouseDZ()/20;

		if( view.y < -89.9f ) view.y = -89.9f;
		if( view.y > 89.9f )  view.y = 89.9f;
		if( view.z < 2.0f )	view.z = 2.0f;

		cam.SetToPosition( Vector3DF(0,0,0) );
		cam.SetFromAngles( view );
		axisView.x = view.x;
		axisView.y = view.y;
	}

	// Toggle Console on '`' key press
	if ( GameX.GetKeyPress(KEY_TILDE) )
		console->Toggle();

	if ( GameX.GetKeyPress(KEY_0) )
		graphsDisabled = !graphsDisabled;

	if (!graphsDisabled)
	{
		if ( GameX.GetKeyPress(KEY_1) )
			accGraph[0]->Toggle();
		if ( GameX.GetKeyPress(KEY_2) )
			accGraph[1]->Toggle();
		if ( GameX.GetKeyPress(KEY_3) )
			accGraph[2]->Toggle();
		if ( GameX.GetKeyPress(KEY_Q) )
			velGraph[0]->Toggle();
		if ( GameX.GetKeyPress(KEY_W) )
			velGraph[1]->Toggle();
		if ( GameX.GetKeyPress(KEY_E) )
			velGraph[2]->Toggle();
		if ( GameX.GetKeyPress(KEY_A) )
			posGraph[0]->Toggle();
		if ( GameX.GetKeyPress(KEY_S) )
			posGraph[1]->Toggle();
		if ( GameX.GetKeyPress(KEY_D) )
			posGraph[2]->Toggle();
		if ( GameX.GetKeyPress(KEY_Z) )
			angGraph[0]->Toggle();
		if ( GameX.GetKeyPress(KEY_X) )
			angGraph[1]->Toggle();
		if ( GameX.GetKeyPress(KEY_C) )
			angGraph[2]->Toggle();

		if ( GameX.GetKeyPress(KEY_9) )
		{
			showAllGraphs = !showAllGraphs;
			accGraph[0]->SetShowGraph(showAllGraphs);
			accGraph[1]->SetShowGraph(showAllGraphs);
			accGraph[2]->SetShowGraph(showAllGraphs);
			velGraph[0]->SetShowGraph(showAllGraphs);
			velGraph[1]->SetShowGraph(showAllGraphs);
			velGraph[2]->SetShowGraph(showAllGraphs);
			posGraph[0]->SetShowGraph(showAllGraphs);
			posGraph[1]->SetShowGraph(showAllGraphs);
			posGraph[2]->SetShowGraph(showAllGraphs);
			angGraph[0]->SetShowGraph(showAllGraphs);
			angGraph[1]->SetShowGraph(showAllGraphs);
			angGraph[2]->SetShowGraph(showAllGraphs);
		}
	}
}

void drawAxis ()
{
	GameX.DrawLine(ColorX(150,150,150), 0, height, width, height);
	GameX.DrawLine(ColorX(150,150,150), width, 0, width, height);

	cam.SetWindow ( width-100, 0, width, 100 );
	cam.SetFromAngles( axisView );
	
	GameX.DrawLine3D (ColorX(255,0,0), Vector3DF(0,0,0), Vector3DF(24,0,0));
	GameX.DrawLine3D (ColorX(0,255,0), Vector3DF(0,0,0), Vector3DF(0,24,0));
	GameX.DrawLine3D (ColorX(0,0,255), Vector3DF(0,0,0), Vector3DF(0,0,24));

	float x, y, z;
	cam.Project( Vector3DF(24,0,0), x, y, z);
	GameX.DrawText(x-5, y-5, txt_x);
	cam.Project( Vector3DF(0,24,0), x, y, z);
	GameX.DrawText(x-5, y-5, txt_y);
	cam.Project( Vector3DF(0,0,24), x, y, z);
	GameX.DrawText(x-5, y-5, txt_z);

	cam.SetWindow (0, 0, width, height);
	cam.SetFromAngles( view );

	sprintf(buffer, "x:%5.3f, y:%5.3f, z:%5.3f", wrist->pos.x, wrist->pos.y, wrist->pos.z);
	GameX.DrawText(5,5,buffer);	
	sprintf(buffer, "vx:%5.3f, vy:%5.3f, vz:%5.3f", wrist->vel.x, wrist->vel.y, wrist->vel.z);
	GameX.DrawText(5,20,buffer);	
	sprintf(buffer, "ax:%5.3f, ay:%5.3f, az:%5.3f", wrist->x_avg, wrist->acc.y, wrist->acc.z);
	GameX.DrawText(5,35,buffer);
	sprintf(buffer, "phi:%5.3f, psi:%5.3f , theta:%5.3f", wrist->phi, wrist->psi, wrist->theta);
	GameX.DrawText(5,50,buffer);
}

void GameDraw (void)
{
	GameX.ClearScreen();
	
	// Draw Gray Background
	GameX.SetView(0,0,width,height);
	GameX.FillScreen (ColorX(100,100,100));
	GameX.ResetView();

	GameX.Begin3DScene ();

	body->Draw();

	//wrist->Draw();
	//elbow->Draw();
	shoulder->Draw();

	if (!graphsDisabled)
	{
		accGraph[0]->Draw();
		accGraph[1]->Draw();
		accGraph[2]->Draw();
		velGraph[0]->Draw();
		velGraph[1]->Draw();
		velGraph[2]->Draw();
		posGraph[0]->Draw();
		posGraph[1]->Draw();
		posGraph[2]->Draw();
		angGraph[0]->Draw();
		angGraph[1]->Draw();
		angGraph[2]->Draw();
	}

	startLogButton->Draw();
	saveLogButton->Draw();
	runLogButton->Draw();
	resetButton->Draw();

	if (logSerialToFile)
		GameX.DrawText(10, 700, "Logging...");

	if (readFromLogFile)
		GameX.DrawText(10, 700, "Reading from log file...");

	if (doneReadingLog)
		GameX.DrawText(10, 700, "Done Reading from log file!");
	
	drawAxis ();

	GameX.End3DScene ();

	console->Draw();	
}