#include "Wrist.hpp"
#include "Console.hpp"

extern Console *console;


float yn_1, xn_1, R;
Wrist::Wrist() : Point()
{
}

Wrist::Wrist(Vector3DF _pos, int _dist, Point *_parent, CameraX *_cam) : Point (_pos, _dist, _parent, _cam)
{
	//offset.Set( 189, 190, 195 );
	offset.Set( 187, 187, 187 );
	gravity.Set( 31.5, 32, 30.5 );

	// set all initial samples to zero
	for( int i=0; i<NUM_SAMPLES; i++ )
		avg_vector[i] = 0;

	oldest_ptr = 0;
	x_avg = 0;
	zero_cnt = 0;
	yn_1 = 0;
	xn_1 = 0;
	R = .1;
	dist = 120; 

	vel_up = false;
	vel_down = false;
	
}



void Wrist::Run( Vector3DF rawAcc )
{
	// Get the acceleration values by subtracting the DC offset
	acc = (rawAcc-OFFSET);

	// compute angles
	phi = atan2( acc.y, acc.z );
	theta = atan2( acc.z, acc.x );

	// lowpass filter the acceleration
	LPFilter( acc );

	// filter x-axis acc further for numerical integration
	filter_X();

	//psi += acc.x;

	// update position based upon angles
	pos.x = dist*cos( psi-PI/2 );
	pos.y = dist*cos( phi ) + dist*sin( psi );
	pos.z = dist*sin( phi );

	// set your position based upon your parent's
	pos = parent->pos + pos;

	// make sure you're always d away from your parent
	Vector3DF d = pos - parent->pos;
	pos = parent->pos + d*(dist/d.Length());

}



void Wrist::filter_X(){

	//---------- Do x-axis DC-blocker ----------//
	// implement y[n] = x[n] - x[n-1] + R*y[n-1]
	float temp = acc.x;		// x[n] value
	acc.x = temp - xn_1 + R*yn_1;

	// update y[n-1],x[n-1]
	yn_1 = acc.x;
	xn_1 = temp;
	//--------- End DC blocker ----------//

	//---------- Do Moving average filter on acceleration ----------//
	// update the avg acceleration
	//  add the newest sample, subtract the oldest sample
	x_avg -= avg_vector[oldest_ptr]/NUM_SAMPLES;
	avg_vector[oldest_ptr] = acc.x;
	x_avg += acc.x/NUM_SAMPLES;

	// put the newest sample in the oldest sample spot
	avg_vector[oldest_ptr] = acc.x;

	// the next element is now the oldest sample
	oldest_ptr++;
	if( oldest_ptr == NUM_SAMPLES )
		oldest_ptr = 0;
	//---------- End Filtering acceleration ----------//

	//---------- Magnitude shape the velocity and Do Numerical Integration ----------// 
	if( x_avg >= 0.f )
		vel.x += log( x_avg+1 )/10;
	if( x_avg < 0.0f )
		vel.x -= log( -1*x_avg + 1 )/10;

	if( -0.01f < vel.x && vel.x < 0.01f )
		vel.x = 0;
	//---------- End Magnitude shape velocity, Num Integration ----------//

	//---------- Damp the velocity ----------//
	if( x_avg > -.4 && x_avg < .4 ){
		vel.x *= .1;
		zero_cnt++;
	}
	//---------- End Velocity damping ----------//

	//---------- Debounce velocity ----------//
	if( vel.x < -.5f && !vel_up ){
		vel_down = true;
	}else if( vel.x > .5f && !vel_down ){
		vel_up = true;
	}else if( zero_cnt >= 3 ){
		vel_up = false;
		vel_down = false;
		zero_cnt = 0;
	}

	if( vel.x > 0 && vel_down )
		vel.x = 0;
	if( vel.x < 0 && vel_up )
		vel.x = 0;



	// ---------- numerically integrate the velocity to extract angular position ----------//
	//pos.x += vel.x;
	psi += vel.x/dist;

	// put bounds on the angle
	if( psi <= -PI/2 )
		psi = -PI/2;

	if( psi >= PI/2 )
		psi = PI/2;
}

void Wrist::Draw()
{
	// Draw the arm
	GameX.DrawLine3D( ColorX(150,150,150), pos, parent->pos );

	// display the point
	GameX.DrawImage3D ( &image, pos );

	// display wrist orientation
	GameX.DrawLine3D( ColorX(255,0,0), pos, Vector3DF( pos.x-10*cos(theta), pos.y, pos.z+10*sin(theta) ) );

}

//---------- Set everything back to zero ----------//
void Wrist::Reset()
{
	Point::Reset();
	x_avg = 0;
	zero_cnt = 0;
	oldest_ptr = 0;
	for (int i = 0; i < NUM_SAMPLES; i++)
		avg_vector[i] = 0;
}
